//
//  QAAppDelegate.m
//  SceneKit Demo
//
//  Created by Daniel Dönigus on 08.11.12.
//  Copyright (c) 2012 Quality Applications. All rights reserved.
//

#import "QAAppDelegate.h"

#define OBJECT_DISTANCE 1.5

@implementation QAAppDelegate

//Hmmm, somehow this small presentation really got out of control in terms of lines of code...
//But everything had to be animated and selectable... :-/

#pragma mark - initialization

- (void) applicationDidFinishLaunching:(NSNotification *)aNotification {
    [self changeColorOfContentView];
    [self retrieveBouncingAnimation];
    [self createScene];
    [self insertFloor];
    [self addAllCreatableObjects];
    [self buildScene];
    [self animateSpot];
    [self buildFadeOutAnimation];
    [self addTextureObjects];
    [self addColorObjects];
    
    self.removedNodes = [NSMutableArray array];
}

- (void) changeColorOfContentView {
    //Needed to show a thin brown line between the scene view and control view
    ((NSView*)self.window.contentView).superview.wantsLayer = YES;
    CALayer* layer = [CALayer layer];
    [layer setBackgroundColor:[NSColor colorWithCalibratedRed:0.34375 green:0.265625 blue:0.171875 alpha:1.0].CGColor];
    ((NSView*)self.window.contentView).layer = layer;
}

- (void) createScene {
    //Create scene and assign to scene kit view and delegate
    self.sceneView.scene = [SCNScene scene];
    self.scene = self.sceneView.scene;
    SCNNode* rootNode = self.scene.rootNode;
    
    //Create camera node and add to scene graph
    SCNNode* cameraNode = [SCNNode node];
    cameraNode.position = SCNVector3Make(0.0, -8.0, 3.0);
    cameraNode.rotation = SCNVector4Make(1.0, 0.0, 0.0, M_PI/8*3);
    cameraNode.camera = [SCNCamera camera];
    //Add node to scene graph
    [rootNode addChildNode:cameraNode];
    
    //Create node with light and add to scene graph
    SCNNode* lightNode = [SCNNode node];
    lightNode.position = SCNVector3Make(0.0, -6.0, 10.0);
    lightNode.rotation = SCNVector4Make(1.0, 0.0, 0.0, 0.0);
    //Create light and parameters
    SCNLight* light = [SCNLight light];
    lightNode.light = light;
    light.type = SCNLightTypeSpot;
    light.color = [NSColor lightGrayColor];
    [light setAttribute:@200.0 forKey:SCNLightSpotOuterAngleKey];
    //Add node to scene graph
    [rootNode addChildNode:lightNode];
    
    //Create an additional spot to illuminate the selected object
    SCNNode* selectedObjectLightNode = [SCNNode node];
    selectedObjectLightNode.rotation = SCNVector4Make(1.0, 0.0, 0.0, M_PI/4);
    //Create light object and parameters
    SCNLight* selectedObjectLight = [SCNLight light];
    selectedObjectLight.type = SCNLightTypeSpot;
    [selectedObjectLight setAttribute:@15.0 forKey:SCNLightSpotOuterAngleKey];
    [selectedObjectLight setAttribute:@1.0 forKey:SCNLightAttenuationFalloffExponentKey];
    selectedObjectLight.color = [NSColor darkGrayColor];
    selectedObjectLightNode.light = selectedObjectLight;
    self.spotLightNode = selectedObjectLightNode;
    //Add node to scene graph
    [rootNode addChildNode:selectedObjectLightNode];
}

- (void) addObject:(SCNGeometry*) object atHeight:(float) height andName:(NSString*) name atIndex:(unsigned long) index {
    //Create a node
    SCNNode* node = [SCNNode node];
    //Rendering order has to be set to render transparency correctly
    node.renderingOrder = 1;
    //Add geometry to node
    node.geometry = object;
    //Add node to scene graph
    [[self.scene rootNode] addChildNode:node];
    //Create new material
    SCNMaterial* material = [SCNMaterial material];
    object.firstMaterial = material;
    
    if ([name isEqualTo:@"Krawatte"]) {
        //The tie-logo is red
        material.diffuse.contents = [NSColor redColor];
    } else {
        //all the other objects are blue
        material.diffuse.contents = [NSColor blueColor];
    }
    
    //Set the node transformation
    node.rotation = SCNVector4Make(1.0, 0.0, 0.0,M_PI/2);
    node.position = SCNVector3Make(0.0, 0.0, height);
    
    //For the tie we need to load additional geometry. I know this is really ugly, but I'm a little lazy at the moment.
    //This only one special case for a small presentation, so please live with it.
    if ([name isEqualTo:@"Krawatte"]) {
        //Loading tie scene
        NSURL          *sceneURL        = [[NSBundle mainBundle] URLForResource:@"krawatte" withExtension:@"dae"];
        SCNSceneSource *sceneSource     = [SCNSceneSource sceneSourceWithURL:sceneURL options:nil];
        
        //Retrieving geometry parts
        SCNGeometry* krawatte1 = [sceneSource entryWithIdentifier:@"Mesh-mesh" withClass:[SCNGeometry class]];
        SCNGeometry* krawatte2 = [sceneSource entryWithIdentifier:@"Mesh_001-mesh" withClass:[SCNGeometry class]];
        
        //Create node, position and material for first part
        SCNNode* subNode1 = [SCNNode node];
        subNode1.position = SCNVector3Make(0.0, -0.7*0.145, 0.0);
        subNode1.geometry = krawatte1;
        krawatte1.firstMaterial.diffuse.contents = [NSColor colorWithCalibratedRed:0.2 green:0.2 blue:0.2 alpha:1.0];
        
        //Create node, position and material for first part
        SCNNode* subNode2 = [SCNNode node];
        subNode2.position = SCNVector3Make(0.0, 1.85*0.145, 0.0);
        subNode2.geometry = krawatte2;
        krawatte2.firstMaterial.diffuse.contents = [NSColor colorWithCalibratedRed:0.2 green:0.2 blue:0.2 alpha:1.0];

        //Addd nodes as subnodes of actual tie-node
        [node addChildNode:subNode1];
        [node addChildNode:subNode2];
    }
    
    //Set the name of the node
    node.name = name;
    
    //Insert node to array of shown nodes
    [self.nodes insertObject:node atIndex:index];
}

- (void) addObject:(SCNGeometry*) object atHeight:(float) height andName:(NSString*) name {
    unsigned long count = self.nodes.count;
    [self addObject:object atHeight:height andName:name atIndex:count];
}

- (NSDictionary*) dictionrayForCreatableObject:(SCNGeometry*) object withHeight:(NSNumber*) height andName:(NSString*) name {
    //Convenience method to create a NSDictionary containing tome objects
    return [NSDictionary dictionaryWithObjectsAndKeys:object, @"Geometry", height, @"Height", name, @"Name", nil];
}

- (void) addAllCreatableObjects {
    //Add several geometries to an array of creatable objects
    NSMutableArray* creatableObjects = [NSMutableArray array];
    
    SCNCone* cone = [SCNCone coneWithTopRadius:0 bottomRadius:0.4 height:1];
    //For each creatable object the geometry, height and a name is stored
    [creatableObjects addObject:[self dictionrayForCreatableObject:cone withHeight:@0.5 andName: @"Cone"]];
    
    SCNBox* box = [SCNBox boxWithWidth:1 height:1 length:1 chamferRadius:0];
    [creatableObjects addObject:[self dictionrayForCreatableObject:box withHeight:@0.5 andName: @"Box"]];
    
    SCNBox* boxChamfer = [SCNBox boxWithWidth:1 height:1 length:1 chamferRadius:0.1];
    [creatableObjects addObject:[self dictionrayForCreatableObject:boxChamfer withHeight:@0.5 andName: @"Box Chamfer"]];
    
    SCNCapsule* capsule = [SCNCapsule capsuleWithCapRadius:0.2 height:1];
    [creatableObjects addObject:[self dictionrayForCreatableObject:capsule withHeight:@0.5 andName: @"Capsule"]];
    
    SCNCylinder* cylinder = [SCNCylinder cylinderWithRadius:0.3 height:1];
    [creatableObjects addObject:[self dictionrayForCreatableObject:cylinder withHeight:@0.5 andName: @"Cylinder"]];
    
    SCNPyramid* pyramid = [SCNPyramid pyramidWithWidth:1 height:1 length:1];
    [creatableObjects addObject:[self dictionrayForCreatableObject:pyramid withHeight:@0.0 andName: @"Pyramid"]];
    
    //A text geometry object.
    SCNText* text = [SCNText textWithString:@"Macoun" extrusionDepth:0.1];
    text.font = [NSFont fontWithName:@"Helvetica" size:0.4];
    CGSize size = text.textSize;
    text.containerFrame = CGRectMake(-size.width/2, 0, size.width, 0);
    [creatableObjects addObject:[self dictionrayForCreatableObject:text withHeight:@0.0 andName: @"Text"]];
    
    SCNSphere* sphere = [SCNSphere sphereWithRadius:0.5];
    [creatableObjects addObject:[self dictionrayForCreatableObject:sphere withHeight:@0.5 andName: @"Sphere"]];
    
    SCNTorus* torus = [SCNTorus torusWithRingRadius:0.4 pipeRadius:0.1];
    [creatableObjects addObject:[self dictionrayForCreatableObject:torus withHeight:@0.1 andName: @"Torus"]];
    
    SCNTube* tube = [SCNTube tubeWithInnerRadius:0.2 outerRadius:0.4 height:1];
    [creatableObjects addObject:[self dictionrayForCreatableObject:tube withHeight:@0.5 andName: @"Tube"]];
    
    //Retreiving the Macoun-Logo from logo.dae scene.
    NSURL          *sceneURL        = [[NSBundle mainBundle] URLForResource:@"logo" withExtension:@"dae"];
    SCNSceneSource *sceneSource     = [SCNSceneSource sceneSourceWithURL:sceneURL options:nil];
    SCNGeometry* logo = [sceneSource entryWithIdentifier:@"Mesh-mesh" withClass:[SCNGeometry class]];
    [creatableObjects addObject:[self dictionrayForCreatableObject:logo withHeight:@0.5 andName: @"Logo"]];
    
    //Retreiving the tie from krawatte.dae scene.
    NSURL          *sceneURLK        = [[NSBundle mainBundle] URLForResource:@"krawatte" withExtension:@"dae"];
    SCNSceneSource *sceneSourceK     = [SCNSceneSource sceneSourceWithURL:sceneURLK options:nil];
    SCNGeometry* krawatte = [sceneSourceK entryWithIdentifier:@"Mesh_002-mesh" withClass:[SCNGeometry class]];
    [creatableObjects addObject:[self dictionrayForCreatableObject:krawatte withHeight:@0.5 andName: @"Krawatte"]];
    
    self.creatableObjects = creatableObjects;
}

- (void) insertFloor {
    //Insert floor to mirror scene
    SCNNode* floorNode = [SCNNode node];
    floorNode.rotation = SCNVector4Make(1.0, 0.0, 0.0, M_PI_2);
    SCNFloor* floor = [SCNFloor floor];
    //Rendering order has to be set to render transparency correctly
    floorNode.renderingOrder = -1;
    floorNode.geometry = floor;
    //Add floor node to scene graph
    [[self.scene rootNode] addChildNode:floorNode];
    
    //Set wood texture as material for floor
    SCNMaterial* material = floor.firstMaterial;
    NSString * texturePath = [[NSBundle mainBundle] pathForResource:@"wood" ofType:@"jpg"];
    NSImage* image = [[NSImage alloc] initWithContentsOfFile:texturePath];
    material.diffuse.contents = image;
    material.diffuse.contentsTransform = CATransform3DScale(CATransform3DIdentity, 5.0, 10.0, 1);
}

- (void) buildScene {
    self.nodes = [NSMutableArray array];
    
    //Addd all creatable objects to scene
    for (NSDictionary* dict in self.creatableObjects) {
        SCNGeometry* geometry = [[dict objectForKey:@"Geometry"] copy];
        float height = [[dict objectForKey:@"Height"] floatValue];
        NSString* name = [dict objectForKey:@"Name"];
        
        //except the text object
        if (![name isEqualToString:@"Text"]) {
            [self addObject:geometry atHeight:height andName:name];
        }
    }
    
    //Select the sphere node
    [self selectNode: [self.nodes objectAtIndex:6]];
    //Repostion all nodes
    [self setPositionForElementsExcludeSelected:NO];
    //and the spot on the selected node
    [self calculateSpotPosition];
}

#pragma mark - Create and build explicit animations

- (void) retrieveBouncingAnimation {
    //Retreiving animation from DAE-File
    NSURL          *sceneURL        = [[NSBundle mainBundle] URLForResource:@"bouncing" withExtension:@"dae"];
    SCNSceneSource *sceneSource     = [SCNSceneSource sceneSourceWithURL:sceneURL options:nil];
    CAAnimation* positionAnimation = [sceneSource entryWithIdentifier:@"Cube_location_Z" withClass:[CAAnimation class]];
    CAAnimation* scaleAnimation = [sceneSource entryWithIdentifier:@"Cube_scale_Y" withClass:[CAAnimation class]];
    
    //The animation consists of two animation parts, so we have to create an animation group
    self.insertAnimation = [CAAnimationGroup animation];
    NSArray* animations = [NSArray arrayWithObjects:positionAnimation, scaleAnimation, nil];
    self.insertAnimation.animations = animations;
    self.insertAnimation.duration = scaleAnimation.duration;
}

- (void) buildFadeOutAnimation {
    //Building an animation group manually
    self.fadeOutAnimation = [CAAnimationGroup animation];
    self.fadeOutAnimation.duration = 2.0;
    
    //First part of the animation setting opacity to 0.0
    CABasicAnimation* opatityAnimation = [CABasicAnimation animationWithKeyPath:@"opacity"];
    opatityAnimation.toValue = [NSNumber numberWithFloat:0.0];
    
    //Second part of the animation setting z-position to 3.0
    CABasicAnimation* positionAnimation = [CABasicAnimation animationWithKeyPath:@"position.z"];
    positionAnimation.toValue = [NSNumber numberWithFloat:3.0];
    
    //Adding both parts of the animation to animation group
    NSArray* animations = [NSArray arrayWithObjects:opatityAnimation, positionAnimation, nil];
    self.fadeOutAnimation.animations = animations;
    self.fadeOutAnimation.delegate = self;
}

#pragma mark - Insert texture objects to scene

- (SCNNode*) addTextureObjectAtPosition:(SCNVector3) position
                            withTexture:(NSString*) textureName
                               andAngle:(float) angle{
    //Create transformation node with rotation around z-axis, because the geometry is rotated
    SCNNode* transformNode = [SCNNode node];
    transformNode.rotation = SCNVector4Make(0.0, 0.0, 1.0, -M_PI_2);
    
    //We need an additional node to do the repositioning when a texture has been selected
    SCNNode* node = [SCNNode node];
    
    //Retreive geometry from papier.dae scene
    NSURL          *sceneURL        = [[NSBundle mainBundle] URLForResource:@"papier" withExtension:@"dae"];
    SCNSceneSource *sceneSource     = [SCNSceneSource sceneSourceWithURL:sceneURL options:nil];
    SCNGeometry* geometry = [sceneSource entryWithIdentifier:@"Plane-mesh" withClass:[SCNGeometry class]];
    
    //Set transformation for mid node
    node.geometry = geometry;
    node.position = position;
    node.rotation = SCNVector4Make(0.0, 0.0, 1.0, angle);
    node.scale = SCNVector3Make(10.0, 10.0, 4.0);
    
    //Load texture image and assign to material
    NSString * texturePath = [[NSBundle mainBundle] pathForResource:textureName ofType:@"jpg"];
    NSImage* image = [[NSImage alloc] initWithContentsOfFile:texturePath];
    SCNMaterial* material = [SCNMaterial material];
    material.diffuse.contents = image;
    material.doubleSided = YES;
    geometry.firstMaterial = material;
    
    //Add mid node to parent
    [transformNode addChildNode:node];
    
    //Add parent node to scene graph
    [self.scene.rootNode addChildNode:transformNode];
    
    return transformNode;
}

- (void) addTextureObjects {
    //Insert geometry to indicate texture selection
    self.textureJeansNode = [self addTextureObjectAtPosition:SCNVector3Make(4.0, 3.0, 0.04)
                                                 withTexture:@"jeans"
                                                    andAngle:M_PI/9];
    self.textureWoodNode = [self addTextureObjectAtPosition:SCNVector3Make(4.2, 2.5, 0.04)
                                                withTexture:@"wood"
                                                   andAngle:M_PI/5];
    self.textureEarthNode = [self addTextureObjectAtPosition:SCNVector3Make(4.4, 2.0, 0.04)
                                                 withTexture:@"earthmap1k"
                                                    andAngle:M_PI/2];
}

#pragma mark - Insert color objects to scene

- (SCNNode*) addColorObjectAtPosition:(SCNVector3) position {
    //Create node
    SCNNode* transformNode = [SCNNode node];
    transformNode.position = position;
    
    //Create sphere as geometry for node
    SCNGeometry* geometry = [SCNSphere sphereWithRadius:0.2];
    transformNode.geometry = geometry;

    //Add node to scene graph
    [self.scene.rootNode addChildNode:transformNode];
    
    return transformNode;
}

- (void) addColorObjects {
    //Add sphere to set a preselected color to selected node
    self.materialRedMatteNode = [self addColorObjectAtPosition:SCNVector3Make(-2.8, -4.0, 0.2)];
    //Set the color to sphere
    [self applyRedMatteMaterialToNode:self.materialRedMatteNode];
    self.materialBlueGlossyNode = [self addColorObjectAtPosition:SCNVector3Make(-2.2, -4.2, 0.2)];
    [self applyBlueGlossyMaterialToNode:self.materialBlueGlossyNode];
    self.materialWhiteMatteNode = [self addColorObjectAtPosition:SCNVector3Make(-1.6, -4.4, 0.2)];
    [self applyWhiteMatteMaterialToNode:self.materialWhiteMatteNode];
}

#pragma mark - Insertion and removal of objects

- (IBAction)insertAdditionalObject:(id)sender {
    //Do nothing if nothing has been selected
    if (self.selectedInsertGeometry == nil) {
        return;
    }
    
    NSUInteger indexOfSelectedNode = [self.nodes indexOfObject:self.selectedNode];
    
    SCNGeometry* geometry;
    //Needed to insert object add correct height
    float height = [[self.selectedInsertGeometry objectForKey:@"Height"] floatValue];
    //If is needed to recreate the text, because it could not be created in applicationDidFinishLaunching
    NSString* name = [self.selectedInsertGeometry objectForKey:@"Name"];
    if ([name isEqualToString:@"Text"]) {
        SCNText* text = [SCNText textWithString:@"Macoun" extrusionDepth:0.1];
        text.font = [NSFont fontWithName:@"Helvetica" size:0.4];
        CGSize size = text.textSize;
        text.containerFrame = CGRectMake(-size.width/2, 0, size.width, 0);
        geometry = text;
    } else {
        geometry = [[self.selectedInsertGeometry objectForKey:@"Geometry"] copy];
    }
    
    [self addObject:geometry atHeight:height andName: name atIndex:indexOfSelectedNode];
    //Select the new inserted Object
    [self selectNode: [self.nodes objectAtIndex:indexOfSelectedNode]];
    //Reposition all nodes except the new object. 
    [self setPositionForElementsExcludeSelected:YES];
    
    //We need to set the position of the new object without animation, because the implicit repositioning and the
    //explicit insertion animation simultaneously would look weird.
    long size = self.nodes.count;
    float radius = OBJECT_DISTANCE/(2.0*sin(M_PI/size));
    self.selectedNode.position = SCNVector3Make(0.0, -radius, self.selectedNode.position.z);
    
    //Add explicit animation to selected node
    [self.selectedNode addAnimation:self.insertAnimation forKey:nil];
    //and reposition the spot on selected object
    [self calculateSpotPosition];
}

- (IBAction)removeSelectedObject:(id)sender {
    NSUInteger indexOfSelectedNode = [self.nodes indexOfObject:self.selectedNode];
    //Remove node from array of nodes
    [self.nodes removeObjectAtIndex:indexOfSelectedNode];
    //Add explicit animation to selected node
    [self.selectedNode addAnimation:self.fadeOutAnimation forKey:nil];
    //Add selected node to an array to delete it just after the animation finished
    [self.removedNodes addObject:self.selectedNode];
    //Select next node
    [self selectNode: [self.nodes objectAtIndex:indexOfSelectedNode%self.nodes.count]];
    //Reset position of remaining nodes 
    [self setPositionForElementsExcludeSelected:NO];
    //and reposition the spot on selected object
    [self calculateSpotPosition];
}

#pragma mark - Helper methods

- (void) calculateSpotPosition {
    //Number of objects in circle
    long size = self.nodes.count;
    //Radius dependent on number of objects
    float radius = OBJECT_DISTANCE/(2.0*sin(M_PI/size));
    
    //Change position of spot on selected object dependent on radius
    [SCNTransaction begin];
    [SCNTransaction setAnimationDuration:2.0];
    self.spotLightNode.position = SCNVector3Make(0.0, -radius-5.0, 5.0);
    [SCNTransaction commit];
}

- (void)animationDidStop:(CAAnimation *)theAnimation finished:(BOOL)flag {
    //Remove the node after the remove-animation is finished
    if (theAnimation == self.fadeOutAnimation) {
        if ([self.removedNodes count] > 0) {
            SCNNode* node = [self.removedNodes objectAtIndex:0];
            [self.removedNodes removeObjectAtIndex:0];
            [node removeFromParentNode];
        }
    }
}

- (void) setPositionForElementsExcludeSelected:(BOOL) exclude {
    //Number of objects in circle
    long size = self.nodes.count;
    //Radius dependent on number of objects
    float radius = OBJECT_DISTANCE/(2.0*sin(M_PI/size));
    
    NSUInteger indexOfSelectedNode = [self.nodes indexOfObject:self.selectedNode];
    
    //Change position of all objects in implicit animation
    [SCNTransaction begin];
    [SCNTransaction setAnimationDuration:2.0];
    int startIndex = 0;
    if (exclude) {
        startIndex = 1;
    }
    for (int i = startIndex; i < [self.nodes count]; i++) {
        SCNNode* node = [self.nodes objectAtIndex:(indexOfSelectedNode+i)%[self.nodes count]];
        //Calculating position with some trigonometry
        node.position = SCNVector3Make(sin(M_PI*2*i/size)*radius,-cos(M_PI*2*i/size)*radius,node.position.z);
    }
    [SCNTransaction commit];
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context {
    //The diffuse contents has been changed via color well.
    //Deselect the currently selected material node
    [self moveSelectedMaterialNodeDown];
    
    //Change the name of the material to indicate that it isn't one of the presettings
    self.selectedNode.geometry.firstMaterial.name = nil;
}

- (void) selectNode:(SCNNode*) node {
    //Add delegate as observer for diffuse contents, so we receive changes to the color done via color well
    [node.geometry.firstMaterial addObserver:self forKeyPath:@"diffuse.contents" options:NSKeyValueObservingOptionOld|NSKeyValueObservingOptionNew  context:NULL];

    //Remove delegate as observer to currently selected object, before selecting the new node
    [self.selectedNode.geometry.firstMaterial removeObserver:self forKeyPath:@"diffuse.contents"];
    self.selectedNode = node;
    
    
    //Select correct material node for the new selected node
    [self selectMaterialNode];
}

#pragma mark - Animate spot of selected node

- (void) animateSpot {
    //Animate the spot on the selected object to alternate between big and small
    [SCNTransaction begin];
    [SCNTransaction setAnimationDuration:2.0];
    [self.spotLightNode.light setAttribute:@13.0 forKey:SCNLightSpotOuterAngleKey];
    [SCNTransaction setCompletionBlock:^(void){
        [SCNTransaction begin];
        [SCNTransaction setAnimationDuration:2.0];
        [self.spotLightNode.light setAttribute:@15.0 forKey:SCNLightSpotOuterAngleKey];
        [SCNTransaction setCompletionBlock:^(void){
            [self animateSpot];
        }];
        [SCNTransaction commit];
    }];
    [SCNTransaction commit];
}

#pragma mark - Handle keyboard events

- (void) leftPressed {
    //The user pressed the left arrow. Select the previous object and change position of all objects
    NSUInteger indexOfNewSelectedNode = ([self.nodes indexOfObject:self.selectedNode] - 1 + [self.nodes count]) % [self.nodes count];
    [self selectNode: [self.nodes objectAtIndex:indexOfNewSelectedNode]];
    
    [self setPositionForElementsExcludeSelected:NO];
}

- (void) rightPressed {
    //The user pressed the right arrow. Select the next object and change position of all objects
    NSUInteger indexOfNewSelectedNode = ([self.nodes indexOfObject:self.selectedNode] + 1) % [self.nodes count];
    [self selectNode: [self.nodes objectAtIndex:indexOfNewSelectedNode]];
    
    [self setPositionForElementsExcludeSelected:NO];
}

#pragma mark - Handle mouse events

- (BOOL) handleMousePressIfPressedOnTextureNode:(SCNNode*) node {
    //If the user has pressed on one of the texture object we need doch change the texture of the selected object
    //and return yes to abort object selection
    if (node.parentNode == self.textureEarthNode) {
        [self textureSelectedWithName:@"Earth" imageName:@"earthmap1k" andMaterialNode:self.textureEarthNode];
        return YES;
    }
    if (node.parentNode == self.textureWoodNode) {
        [self textureSelectedWithName:@"Wood" imageName:@"wood" andMaterialNode:self.textureWoodNode];
        return YES;
    }
    if (node.parentNode == self.textureJeansNode) {
        [self textureSelectedWithName:@"Jeans" imageName:@"jeans" andMaterialNode:self.textureJeansNode];
        return YES;
    }
    
    return NO;
}

- (BOOL) handleMousePressIfPressedOnColorNode:(SCNNode*) node {
    //If the user has pressed on one of the colored spheres we need doch change the color of the selected object
    //and return yes to abort object selection
    if (node == self.materialRedMatteNode) {
        [self applyRedMatteMaterialAnimatedToSelectedNode];
        return YES;
    }
    if (node == self.materialBlueGlossyNode) {
        [self applyBlueGlossyMaterialAnimatedToSelectedNode];
        return YES;
    }
    if (node == self.materialWhiteMatteNode) {
        [self applyWhiteMatteMaterialAnimatedToSelectedNode];
        return YES;
    }
    
    return NO;
}

- (void) mousePressedAtPoint:(CGPoint)point {
    //Do a hit test to receive the selected object
    NSArray* hitTestResults = [self.sceneView hitTest:point options: NULL];
    
    if ([hitTestResults count] > 0) {
        SCNNode* node = [[hitTestResults objectAtIndex:0] node];
        
        //Check if the user has pressed on one of the three texture nodes
        if ([self handleMousePressIfPressedOnTextureNode:node]) {
            return;
        }
        
        //Check if the user has pressed on one of the three colored spheres
        if ([self handleMousePressIfPressedOnColorNode:node]) {
            return;
        }
        
        //The user hasn't pressed on one of the material nodes, so we select the node and change the position of
        //all nodes dependently.
        [self selectNode: node];
        [self setPositionForElementsExcludeSelected:NO];
    }
}

#pragma mark - Handle material changes

- (void) selectMaterialNode {
    //Selects the material node dependent on the first material of the user selected node
    [self moveSelectedMaterialNodeDown];
    self.selectedMaterialNode = nil;
    
    if ([self.selectedNode.geometry.firstMaterial.name isEqualTo:@"RedMatte"]) {
        self.selectedMaterialNode = self.materialRedMatteNode;
    }
    if ([self.selectedNode.geometry.firstMaterial.name isEqualTo:@"BlueGlossy"]) {
        self.selectedMaterialNode = self.materialBlueGlossyNode;
    }
    if ([self.selectedNode.geometry.firstMaterial.name isEqualTo:@"WhiteMatte"]) {
        self.selectedMaterialNode = self.materialWhiteMatteNode;
    }
    if ([self.selectedNode.geometry.firstMaterial.name isEqualTo:@"Wood"]) {
        self.selectedMaterialNode = self.textureWoodNode;
    }
    if ([self.selectedNode.geometry.firstMaterial.name isEqualTo:@"Jeans"]) {
        self.selectedMaterialNode = self.textureJeansNode;
    }
    if ([self.selectedNode.geometry.firstMaterial.name isEqualTo:@"Earth"]) {
        self.selectedMaterialNode = self.textureEarthNode;
    }
    
    if (self.selectedMaterialNode != nil) {
        [self moveSelectedMaterialNodeUp];
    }
}

- (void) moveSelectedMaterialNodeDown {
    //Let the selected material node fly down with an implicit animation
    [SCNTransaction begin];
    [SCNTransaction setAnimationDuration:2.0];
    SCNVector3 oldPosition = self.selectedMaterialNode.position;
    self.selectedMaterialNode.position = SCNVector3Make(oldPosition.x, oldPosition.y, oldPosition.z-0.2);
    [SCNTransaction commit];
}

- (void) moveSelectedMaterialNodeUp {
    //Let the selected material node fly up with an implicit animation
    [SCNTransaction begin];
    [SCNTransaction setAnimationDuration:2.0];
    SCNVector3 oldPosition = self.selectedMaterialNode.position;
    self.selectedMaterialNode.position = SCNVector3Make(oldPosition.x, oldPosition.y, oldPosition.z+0.2);
    [SCNTransaction commit];
}

#pragma mark - Handle color selection

- (void) applyRedMatteMaterialToNode:(SCNNode*) node {
    //Change color parameters
    node.geometry.firstMaterial.ambient.contents = [NSColor darkGrayColor];
    node.geometry.firstMaterial.diffuse.contents = [NSColor redColor];
    node.geometry.firstMaterial.specular.contents = [NSColor blackColor];
    node.geometry.firstMaterial.shininess = 0;
    //Also we need to change the name of the material to reselect the material node when the user reselects the
    //geometry node.
    node.geometry.firstMaterial.name = @"RedMatte";
}

- (void) applyRedMatteMaterialAnimatedToSelectedNode {
    //If the material already has been set we don't need to reassign
    if ([self.selectedNode.geometry.firstMaterial.name isEqualTo:@"RedMatte"]) {
        return;
    }
    
    //Change color values in implicit animation
    [SCNTransaction begin];
    [SCNTransaction setAnimationDuration:2.0];
    [self applyRedMatteMaterialToNode:self.selectedNode];
    [SCNTransaction commit];
    
    //Select the material node and move up.
    self.selectedMaterialNode = self.materialRedMatteNode;
    [self moveSelectedMaterialNodeUp];
}

- (IBAction)redMatte:(id)sender {
    [self applyRedMatteMaterialAnimatedToSelectedNode];
}

- (void) applyBlueGlossyMaterialToNode:(SCNNode*) node {
    //Change color parameters
    node.geometry.firstMaterial.ambient.contents = [NSColor darkGrayColor];
    node.geometry.firstMaterial.diffuse.contents = [NSColor blueColor];
    node.geometry.firstMaterial.specular.contents = [NSColor whiteColor];
    node.geometry.firstMaterial.shininess = 1;
    //Also we need to change the name of the material to reselect the material node when the user reselects the
    //geometry node.
    node.geometry.firstMaterial.name = @"BlueGlossy";
}

- (void) applyBlueGlossyMaterialAnimatedToSelectedNode {
    //If the material already has been set we don't need to reassign
    if ([self.selectedNode.geometry.firstMaterial.name isEqualTo:@"BlueGlossy"]) {
        return;
    }
    
    //Change color values in implicit animation
    [SCNTransaction begin];
    [SCNTransaction setAnimationDuration:2.0];
    [self applyBlueGlossyMaterialToNode:self.selectedNode];
    [SCNTransaction commit];
    
    //Select the material node and move up.
    self.selectedMaterialNode = self.materialBlueGlossyNode;
    [self moveSelectedMaterialNodeUp];
}

- (IBAction)blueGlossy:(id)sender {
    [self applyBlueGlossyMaterialAnimatedToSelectedNode];
}

- (void) applyWhiteMatteMaterialToNode:(SCNNode*) node {
    //Change color parameters
    node.geometry.firstMaterial.ambient.contents = [NSColor darkGrayColor];
    node.geometry.firstMaterial.diffuse.contents = [NSColor whiteColor];
    node.geometry.firstMaterial.specular.contents = [NSColor blackColor];
    node.geometry.firstMaterial.shininess = 0;
    //Also we need to change the name of the material to reselect the material node when the user reselects the
    //geometry node.
    node.geometry.firstMaterial.name = @"WhiteMatte";
}

- (void) applyWhiteMatteMaterialAnimatedToSelectedNode {
    //If the material already has been set we don't need to reassign
    if ([self.selectedNode.geometry.firstMaterial.name isEqualTo:@"WhiteMatte"]) {
        return;
    }
    
    //Change color values in implicit animation
    [SCNTransaction begin];
    [SCNTransaction setAnimationDuration:2.0];
    [self applyWhiteMatteMaterialToNode:self.selectedNode];
    [SCNTransaction commit];
    
    //Select the material node and move up.
    self.selectedMaterialNode = self.materialWhiteMatteNode;
    [self moveSelectedMaterialNodeUp];
}

- (IBAction)whiteMatte:(id)sender {
    [self applyWhiteMatteMaterialAnimatedToSelectedNode];
}

#pragma mark - Handle texture selection

- (void) textureSelectedWithName:(NSString*) name imageName:(NSString*) imageName andMaterialNode:(SCNNode*) node {
    //If the material already has been set we don't need to reassign
    if ([self.selectedNode.geometry.firstMaterial.name isEqualTo:name]) {
        return;
    }
    
    //Set texture image as diffuse contents of material
    NSString * texturePath = [[NSBundle mainBundle] pathForResource:imageName ofType:@"jpg"];
    NSImage* image = [[NSImage alloc] initWithContentsOfFile:texturePath];
    self.selectedNode.geometry.firstMaterial.diffuse.contents = image;
    //Also we need to change the name of the material to reselect the material node when the user reselects the
    //geometry node.
    self.selectedNode.geometry.firstMaterial.name = name;
    
    //Select the material node and move up.
    self.selectedMaterialNode = node;
    [self moveSelectedMaterialNodeUp];
}

- (IBAction)textureWood:(id)sender {
    [self textureSelectedWithName:@"Wood" imageName:@"wood" andMaterialNode:self.textureWoodNode];
}

- (IBAction)textureJeans:(id)sender {
    [self textureSelectedWithName:@"Jeans" imageName:@"jeans" andMaterialNode:self.textureJeansNode];
}

- (IBAction)textureEarth:(id)sender {
    [self textureSelectedWithName:@"Earth" imageName:@"earthmap1k" andMaterialNode:self.textureEarthNode];
}

@end
